#!/usr/bin/env perl
# If you're on a Unix-like system, adjust the above as necessary if the script will not run.


# TwinHelix Forums v1.0 Beta 2, updated 22 February 2015
# Copyright (c) 2002-15 Angus Turnbull http://www.twinhelix.com

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details, at:
# https://gnu.org/licenses/agpl.html


# Please consider making a donation for usage of this script, via my website, as I have
# put a lot of time into developing, debugging and documenting it. Support is always appreciated!
# http://www.twinhelix.com/donate/


# Scroll down a page or two to reach the "Configuration" section.



##### LOADED MODULES ######    <- Ensure your server has these.

# The CGI module is required, the CGI::Carp module is only for debugging.
# I've put a couple of security tweaks there for luck.
use CGI 'param';
$CGI::POST_MAX = 65536;
$CGI::DISABLE_UPLOADS = 1;

# You will want to try uncommenting this if you get script errors, it's helpful when debugging.
# I recommend leaving it commented out once you deploy the script online, though.
#use CGI::Carp 'fatalsToBrowser';




##### GLOBAL VARIABLES #####    <- Just skip this.

# If you feel like it, run this script under "strict" conditions.
# Otherwise you can delete these "use strict/use vars" lines, they're not required.
#use strict;
# The prefs, template, and banned vars contain server-side preferences & settings.
#use vars qw(%forum_prefs %forum_template %forum_banned);
# The accounts index (hash indexed by username), data (by UID) and logged-in username/number.
#use vars qw(%user_index @user_data $user_name $user_num %user_priv %user_groups);
# List of available forums/topics/posts, and a reference to the active forum.
#use vars qw(@forum_data @topic_data @post_data $forum_ref);
# The query is the CGI object, and @path_info stores the extra folder data.
# From those we retrieves the 'action', 'forum', 'topic', 'page' and 'keywords' variables.
#use vars qw($query @path_info $act $forum $topic $page $keywords);
# Optional string out HTML output to be run through a GZIP module.
#use vars qw($out_str);

# Initialise our global variables.
%forum_prefs = ();
%forum_template = ();
%forum_banned = ();
%user_index = ();
@user_data = ();
$user_name = '';
$user_num = 0;
%user_priv = ();
%user_groups = ();
@forum_data = ();
@topic_data = ();
@post_data = ();
$forum_ref = undef;
$forum = '';
$query = new CGI;
@path_info = ();
$act = '';
$forum = '';
$topic = '';
$page = 1;
$keywords = '';
$out_str = '';





##### CONFIGURATION #####     <- Edit This! Important!

# Specify the *local* server folder in which you want the forum data to reside.
# This folder must contain the files "FORUMS.TXT" and "TEMPLATE.PL"
# An important point: this is a SERVER path, on the disk of the web server, for
# instance "C:/folder/forumdata" or "/var/forumdata", depending on your server.
#
# I've put in some handy auto-detection here: $cgi_path points to the current folder
# in which the forum script is running, so you can use that to set relative URLs,
# for examples "$cgipath/../../private" if you want.

my $cgi_path = $1 if (($ENV{'SCRIPT_FILENAME'}||$0) =~ m/^(.*)(\\|\/)(.+?)$/);

$forum_prefs{'base_dir'} = "$cgi_path/forumdata";

# Alternatively, if that fails, you can specify it manually like so:
#$forum_prefs{'base_dir'} = "/var/blah/webstuff/forumdata";
#$forum_prefs{'base_dir'} = "C:/Inetpub/forumdata";
# You may have to email your web hosting provider to find a suitable folder.

# WARNING: Whatever you do, you MUST rename the data folder to something other than
# 'forumdata', and ensure it's not findable by a web browser. This default setting is
# insecure, as anyone can predict its location. Preferably, check with your web hosting
# provider if you can store it in a "private" folder somewhere that is not accessible
# remotely. Otherwise, at the very least rename it with some random characters.



# Next, we must specify the pathname of the script on the server, that the browser will see.
# This auto-detect should work, if it doesn't, try uncommenting the manual setting.
# This is used to construct URLs that point back at this script.

$forum_prefs{'path_name'} = $ENV{'SCRIPT_NAME'}||'';
# $forum_prefs{'path_name'} = "/cgi-bin/forum.pl";

# That's all the configuration you need to do!




##### OPTIONAL EXTRAS  #####    <- Edit these if you want, otherwise leave as is.


# You can select the script mode. In Query mode (the default) URIs are like:
#   /cgi-bin/forum.pl?f=forumname&t=123
# Or in Path mode, URIs are constructed as folder paths:
#   /cgi-bin/forum.pl/forumname/123
# Path mode may have some troubles in IIS, leave it disabled/commented if that occurs.
$forum_prefs{'path_mode'} = 1;


# Enables GZIP compression of script output for supported browsers.
# This will speed up downloads, but increase the load on the server.
# Your server must have the "Compress::Zlib" module (it's a standard part of Perl, usually).
use Compress::Zlib;
$forum_prefs{'use_gzip'} = (($ENV{'HTTP_ACCEPT_ENCODING'}||'') =~ m/gzip/ig) ? 1 : 0;


# The user data store location, defaults to within the main forum data folder.
$forum_prefs{'user_dir'} = "$forum_prefs{'base_dir'}/_users";


# The path=xxxx; value returned in the Set-Cookie header. Leave this as-is, or if you're
# using multiple copies of the forum script on a server, perhaps set to "/".
$forum_prefs{'cookie_path'} = $forum_prefs{'path_name'};


# Characters allowed in usernames. If you have an international audience, add some others.
# If you leave it blank, ALL non-control characters are enabled, which may pose risks.
$forum_prefs{'name_chars'} = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789_@!.- ";


# How long login sessions last, in hours. Default is 30 days.
$forum_prefs{'session_length'} = 30 * 24;


# Whether to lock files when writing. This doesn't work on some versions of Windows.
$forum_prefs{'use_flock'} = 1;


# The maximum amount of bytes users may post in each message to a topic.
$forum_prefs{'max_post_len'} = 2000;
# The maximum amount of message bodies that will be searched in a full-text search.
$forum_prefs{'max_search_messages'} = 100;


# Optionally, define some banned useragents, referrers or IP addresses, in case of abuse.
# I've banned some "site downloader" utilities, as once a clueless user points an ill-configured
# one at your site and it gets to the forum script, your bandwidth bill can skyrocket...
$forum_banned{'HTTP_USER_AGENT'} = ["WebZip", "WebCopier", "WebStripper", "Offline Explorer",
 "Mozilla/3.0 (compatible)"];
#$forum_banned{'HTTP_REFERER'} = ["slashdot.org", "yahoo.com"];
#$forum_banned{'REMOTE_ADDR'} = ["12.34.56.78", "10.11.12.23"];


# User account groups and privileges. You must have "guests", "members" and "admins" groups,
# otherwise you can create however many additional groups you want. Privileges are:
# "view" corresponds to viewing pages past the front page.
# "accview" allows viewing user account info pages.
# "post" is obvious.
# "topicset" allows users to lock, pin, and delete entire topics.
# "ownedit" allows that group to edit their own posts (non-guests only).
# "anyedit" allows that group to edit any post.
# "accedit" allows editing others' accounts and account privilege levels.
$user_groups{'guests'} = { 'name'=>'Guests',
 'view'=>1, 'accview'=>1, 'post'=>1, 'topicset'=>0, 'ownedit'=>0, 'anyedit'=>0, 'accedit'=>0 };
$user_groups{'members'} = { 'name'=>'Members',
 'view'=>1, 'accview'=>1, 'post'=>1, 'topicset'=>0, 'ownedit'=>1, 'anyedit'=>0, 'accedit'=>0 };
$user_groups{'admins'} = { 'name'=>'Moderators',
 'view'=>1, 'accview'=>1, 'post'=>1, 'topicset'=>1, 'ownedit'=>1, 'anyedit'=>1, 'accedit'=>1 };
$user_groups{'banned'} = { 'name'=>'Banned',
 'view'=>0, 'accview'=>0, 'post'=>0, 'topicset'=>0, 'ownedit'=>0, 'anyedit'=>0, 'accedit'=>0 };


##### END OF EDITABLE SECTION #####






##### FORUM DATA MANIPULATION FUNCTIONS #####



# Read out information from a passed forum data/index file, and return an array of hashes.
# Optionally pass a second parameter, the number of records to read.
sub readFileData
{
 if (!$_[0]) { return [] }
 my $numrecs = $_[1];
 my $fileopened = 0;

 my @fileLines = ();

 # Suck in the whole file if we're not reading individual lines.
 # Otherwise, the open() call will still be made, just not the reading.
 if (open(FDAT, $_[0]))
 {
  $fileopened = 1;
  if (!$numrecs)
  {
   @fileLines = <FDAT>;
   close(FDAT) or die "Could not close data file: $_[0], $!";
  }
 }

 # Variables to skip any leading blank lines, as blank lines separate records.
 my $lastWasSpace = 1;
 my $line;
 my @fileData = ();
 my $arrIndex = 0;

 while ($line = ($numrecs ? <FDAT> : shift @fileLines))
 {
  # Trim any extra linebreaks (since various OS's can't agree on a text format).
  $line =~ s/[\r\n]//g;
  if ($line =~ m/^([^=]+)=(.*)/)
  {
   $fileData[$arrIndex]{$1} = $2;
   $lastWasSpace = 0;
  }
  else
  {
   # Increment the array index once for each block of spaces.
   $arrIndex++ if (!$lastWasSpace);
   $lastWasSpace = 1;
  }
  # Bust out of here if we've read enough records.
  last if ($numrecs && ($arrIndex == $numrecs));
 }

 if ($numrecs && $fileopened) { close(FDAT) or die "Could not close data file: $_[0], $!" }
 return @fileData;
}


# Writes out to a filename (parameter 1) a referenced array of hashes (parameter 2).
sub writeFileData
{
 if (!$_[0]) { return 0 }

 my $tmpfile = "$_[0].tmp";
 unlink $tmpfile if (-e $tmpfile);
 open(FDAT, ">$tmpfile") or die "Couldn't open file for writing data: $tmpfile, $!";
 # If this lock fails, perhaps import Fcntl and use the proper constants from there...
 flock(FDAT, 2) if ($forum_prefs{'use_flock'});

 my ($lines, $h, $k, $key_found);
 foreach $h (@{$_[1]})
 {
  foreach $k (keys %{$h})
  {
   $key_found = 1;
   $lines .= "$k=$h->{$k}\n";
  }
  $lines .= "\n" if ($h && $key_found);
  $key_found = 0;
 }

 print FDAT $lines;
 flock(FDAT, 8) if ($forum_prefs{'use_flock'});
 close(FDAT) or die "Could not close file after writing data: $_[0], $!";

 # Rename will overwrite any existing file with this new temp file.
 rename($tmpfile, $_[0]) or die "Couldn't overwrite datafile with 'rename': $!";

 return 1;
}


# Loops through a folder full of forum topics, and updates the corresponding index.txt.
# You must pass a forum name, and optionally pass a topic number as 2nd parameter to
# specifically reread the contents of that file (when it's updated).
sub updateIndex
{
 if (!$_[0]) { return [] }

 # See if we already have @topic_data from another function, if not, read it now.
 if ($#topic_data == -1)
 {
  @topic_data = readFileData("$forum_prefs{'base_dir'}/$_[0]/index.txt");
 }

 # A posts counter we increment in the loops, and a signal to write back new data.
 my $total_posts = 0;
 my $reindex = 0;

 # Now, loop through @topic_data and store a new hash of found topic numbers, where
 # the key equals the actual topic number, and the value the @topic_data index.
 my %topic_indexes = ();
 for (my $t = 0; $t <= $#topic_data; $t++)
 {
  if ($topic_data[$t])
  {
   my $topnum = $topic_data[$t]{'topicnum'};
   # Already indexed? It's a duplicate entry! Remove it from the array...
   if (exists $topic_indexes{$topnum})
   {
    splice(@topic_data, $t, 1);
	$reindex = 1;
   }
   $topic_indexes{$topnum} = $t;
  }
 }

 # Read an array of all files in the topic directory.
 my @dirlines = ();
 if (opendir(TOPICSDIR, "$forum_prefs{'base_dir'}/$_[0]"))
 {
  @dirlines = readdir(TOPICSDIR);
  closedir(TOPICSDIR) or die "Could not close topics dir: $forum_prefs{'base_dir'}/$_[0], $!";
 }
 else { return }

 # Loop through @dirlines, finding entries that are numbers followed by .txt
 foreach my $dirfile (@dirlines)
 {
  # Retrieve the topic number from the filename, or loop for other files.
  my $tn;
  if ($dirfile =~ m/^([0-9]+)\.txt$/) { $tn = $1 }
  else { next }

  # If the topic is on disk but not indexed properly, create a new record.
  # Also check if we've been passed a second parameter, to force a re-index of that file.
  if (!$topic_indexes{$tn} || !$topic_data[$topic_indexes{$tn}]{'topicname'} || ($tn == $_[1]))
  {
   my @file_data = readFileData("$forum_prefs{'base_dir'}/$forum/$tn.txt");
   # Record the new index value as the end of the array if required.
   if (!(exists $topic_indexes{$tn})) { $topic_indexes{$tn} = $#topic_data + 1 }
   $topic_data[$topic_indexes{$tn}] = {
    'topicnum' => $tn,
    'topicname' => $file_data[0]{'name'},
    'attr' => $file_data[0]{'attr'},
    'posts' => $#file_data,
    'origauthor' => $file_data[1]{'author'},
    'origtime' => $file_data[1]{'time'},
    'lastauthor' => $file_data[$#file_data]{'author'},
    'lasttime' => $file_data[$#file_data]{'time'}
   };
   $reindex = 1;
  }

  # Increment the post counter for this topic.
  $total_posts += $topic_data[$topic_indexes{$tn}]{'posts'};
  # Now clear the %topic_indexes entry, so we don't delete it from @topic_data later.
  delete $topic_indexes{$tn};
 }

 # Therefore, all remaining entries in %topic_indexes must be topics that are currently
 # indexed, but not present on disk. So, remove them from the @topic_data array.
 foreach my $deltopic (keys %topic_indexes)
 {
  splice(@topic_data, $topic_indexes{$deltopic}, 1);
  $reindex = 1;
 }

 # Finally, sort our @topic_data array by last post time.
 @topic_data = sort { index($b->{'attr'}, 'P')<=>index($a->{'attr'}, 'P') ||
  $b->{'lasttime'}<=>$a->{'lasttime'} } @topic_data;

 # Call setFileData to write out our resulting array of hashes - pass reference to array.
 writeFileData("$forum_prefs{'base_dir'}/$_[0]/index.txt", \@topic_data) if ($reindex);

 # Replace the content of our "data" file for this forum.
 # Since this is so quick I'm doing it even if we're not doing a full reindex.

 # Scan quickly through the listing to find the first non-pinned topic.
 my $last_topic = 0;
 my $last_time = 0;
 my $scan_topic = 0;
 do
 {
  if ($topic_data[$scan_topic]{'lasttime'} > $last_time)
  {
   $last_topic = $scan_topic;
   $last_time = $topic_data[$scan_topic]{'lasttime'};
  }
 } until ($topic_data[$scan_topic++]{'attr'} !~ /P/);

 my $data_info = {
  'posts' => $total_posts,
  'topics' => $#topic_data + 1,
  'lastauthor' => $topic_data[$last_topic]{'lastauthor'},
  'lasttime' => $last_time,
 };
 writeFileData("$forum_prefs{'base_dir'}/$_[0]/data.txt", [$data_info]);
}








##### COMMON FORMATTING COMMANDS ######


# Remove HTML etc. from a string, replacing it with the appropriate entities.
# First param is the string to process, second is true if we allow linebreaks or false if not.
sub trimHTML
{
 $_ = $_[0] || '';
 s/&/&amp;/g;
 s/"/&quot;/g;
 s/</&lt;/g;
 s/>/&gt;/g;
 s/'/&#39;/g;
 # These are used in CGI.pm but might muck up some charsets...?
 # Comment these out if you're having troubles with a non-Latin charset.
 s/\x8b/&#139;/g;
 s/\x9b/&#155;/g;

 # Now handle control chars and linebreaks -- default is remove them all.
 # Even though the hex chars should catch linebreaks, might as well be safe.
 if (!$_[1]) { s/[\x00-\x1F\r\n]//g }
 else
 {
  # Remove all characters but linebreaks and >= 0x20 (that's decimal 32, the spacebar).
  s/[^\r\n\x20-\xFF]//g;
  # When we allow linebreaks, convert them to to <br />.
  s/\r\n/\n/g;
  s/(\r|\n)/<br \/>/g;
 }

 return $_;
}


# URL encoding and decoding, stolen from CGI.pm and the (mistaken!) ActivePerl documentation.
sub url_enc
{
 my $tmp = $_[0];
 $tmp =~ s/([^\w()'*~!.-])/sprintf('%%%02x', ord($1))/eg;
 return $tmp;
}
sub url_dec
{
 my $tmp = $_[0];
 $tmp =~ s/%([A-Fa-f\d]{2})/chr(hex($1))/eg;
 return $tmp;
}

# GMT time formatting. Pass a date (in seconds), and a second parameter to return only the year.
sub gmtTimeFormat
{
 my ($null,$min,$hour,$day,$mon,$year,$null,$null) = gmtime($_[0]);
 my $ampm = ($hour < 12 ? 'AM' : 'PM');
 $min = sprintf("%02d", $min);
 $hour = $hour ? ($hour > 12 ? $hour-12 : $hour) : 12;
 $mon = ('Jan','Feb','Mar','Apr','May','Jun','Jul','Aug','Sep','Oct','Nov','Dec')[$mon];
 $year += 1900;
 return ($_[1] ? "" : "$hour:$min $ampm, ")."$day $mon $year";
}




# Generic "safe" substitution routine that watches for user-inserted substitution markers.
# It changes all occurances of [!, [?, !] or ?] to [~~! etc., to be changed back later.
# Otherwise users can include substitution markers in their posts, and cause havoc.
sub doSub
{
 my ($subName, $newStr) = @_;
 $newStr = '' unless ($newStr);
 $newStr =~ s/\[(!|\?)/\[~~$1/g;
 $newStr =~ s/(!|\?)\]/$1~~\]/g;
 $newStr =~ s/\^/~~CARET/g;
 s/\[!\Q$subName\E!\]/$newStr/g;
}



# This retrieves a template and does some substitutions common to most templates
# The result is returned as $_ to the calling function.
# Pass it a template name and hash reference to an @forum_data entry (like $forum_ref).
sub getTemplate
{
 my ($template_name, $fRef, $is_row_template) = @_;
 $_ = $forum_template{$template_name};

 # Make usergroups available everywhere...
 s/\[\?ISLOGGEDIN(.)(.*?)\1(.*?)\?\]/$user_name ? $2 : $3/seg;

 doSub("HOMEURL", $forum_prefs{'path_name'});
 # Only substitute these if we are passed a valid forum reference.
 if ($fRef)
 {
  doSub("FORUMNAME", $fRef->{'name'});
  doSub("FORUMTEXT", $fRef->{'text'});
  $fRef->{'id'} |= '';
  if ($forum_prefs{'path_mode'})
  {
   doSub("FORUMURL", "$forum_prefs{'path_name'}/$fRef->{'id'}");
   doSub("NEWTOPICURL", "$forum_prefs{'path_name'}/$fRef->{'id'}/_newtopic");
  }
  else
  {
   doSub("FORUMURL", "$forum_prefs{'path_name'}?f=$fRef->{'id'}");
   doSub("NEWTOPICURL", "$forum_prefs{'path_name'}?f=$fRef->{'id'}&act=newtopic");
  }
 }

 # If we're getting a row template, don't do the next substitution (for speed).
 return if $is_row_template;

 # If we've got some topic information loaded, format it.
 if ($#post_data > -1)
 {
  doSub("TOPICNAME", $post_data[0]{'name'});
  doSub("TOPICURL", $forum_prefs{'path_name'}. ($forum_prefs{'path_mode'} ?
   ($forum ? "/$forum".($topic ? "/$topic".($page ? "/p=$page" : "") : "") : "") :
   ($forum ? "?f=$forum".($topic ? ";t=$topic".($page ? ";p=$page" : "") : "") : "") ) );
 }

 # Attempt to pick some sort of sensible login URL and form submittal URL.
 doSub("LOGINURL", $forum_prefs{'path_name'}. ($forum_prefs{'path_mode'} ?
  ($forum ? "/$forum".($topic ? "/$topic".($page ? "/p=$page" : "") : "") : "") :
  ($forum ? "?f=$forum".($topic ? ";t=$topic".($page ? ";p=$page" : "") : "") : "") ) );
 s/\[!LOGINFIELDS!\]/  <input type="hidden" name="a" value="login" \/>
  <input type="hidden" name="f" value="$forum" \/>
  <input type="hidden" name="t" value="$topic" \/>
  <input type="hidden" name="p" value="$page" \/>/g;
 doSub("USERURL", $forum_prefs{'path_name'}.
  ($forum_prefs{'path_mode'} ? "/_user/" : "?a=users&amp;u=").url_enc($user_name));
 doSub("ACCOUNTURL", $forum_prefs{'path_name'}.
  ($forum_prefs{'path_mode'} ? "/_account/" : "?a=account&amp;u=").url_enc($user_name));
 doSub("LOGOUTURL", $forum_prefs{'path_name'}.($forum_prefs{'path_mode'} ? "/_logout" : "?a=logout"));
 doSub("USERNAME", $user_name);

 # The search information and substitutions.
 s/\[\?ISSEARCH(.)(.*?)\1(.*?)\?\]/$act eq 'search' ? $2 : $3/seg;
 doSub("SEARCHURL", $forum_prefs{'path_name'}.($forum_prefs{'path_mode'} ? "/$forum" : "?f=$forum"));
 s/\[!SEARCHFIELDS!\]/qq|
  <input name="a" type="hidden" value="search" \/>
  <input name="f" type="hidden" value="$forum" \/>
 |/eg;
 doSub("SEARCHKEYWORDS", $keywords ? $keywords : "");
}

# Remove our placeholder munged markers and print it.
sub printTemplate
{
 s/\[~~(!|\?)/\[$1/g;
 s/(!|\?)~~\]/$1\]/g;
 s/~~CARET/\^/g;
 forum_out($_);
}




# Creates a nicely comma-separated list of links to pages, pass it page numbers.
# Uses $_ from the calling function, which should be set to a template string.
sub formatPageLinks
{
 my ($min, $current, $max) = @_;
 my ($pageFormat, $separator, $multisep, $pageStr, $temp) = ("", "", "", "");

 # Set the HASPAGES strings appropriately.
 s/\[\?HASPAGES(.)(.*?)\1(.*?)\?\]/($max > 1) ? $2 : $3/seg;

 # Pull the page list formatting out of $template.
 if (/\[\?PAGELIST(.)(.*?)\1(.*?)\1(.*?)\?\]/) { $pageFormat = $2; $separator = $3; $multisep = $4 }

 # If we have several pages, add each link we need onto $pageStr.
 if ($max > 1)
 {
  # Decide which pages to list... for large lists, can't display 100 links!
  # Adjust this if you want to display more or less links to pages.
  my @show_page_nums = ($min..($min+1), ($current-2)..($current+2), ($max-2)..$max);
  my @done_page_nums = ();
  my $last_page_num = $min - 1;
  foreach my $p (@show_page_nums)
  {
   # Avoid doubleups and out-of-bounds errors.
   next if ($done_page_nums[$p] || ($p < $min) || ($p > $max));
   $done_page_nums[$p] = 1;
   # Sub this number into the formatting template, and tack it onto $pageStr;
   $temp = $pageFormat;
   $temp =~ s/\[!PAGENUM!\]/$p/eg;
   if ($p > $min)
   {
    # Display either the specified separator, or an ellipsis for large gaps.
    $pageStr .= ($p == $last_page_num + 1) ? $separator : $multisep;
   }
   $last_page_num = $p;
   $pageStr .= $temp;
  }
 }

 # Now replace the formatting with our formatted list of pages.
 s/\[\?PAGELIST(.*?)\?\]/$pageStr/seg;

 # Do a similar process for the PREVPAGE and NEXTPAGE markers.
 if (/\[\?PREVPAGE(.)(.*?)\1(.*?)\?\]/)
 {
  $temp = ($current > $min) ? $2 : $3;
  $temp =~ s/\[!PAGENUM!\]/$current - 1/eg;
  s/\[\?PREVPAGE(.*?)\?\]/$temp/seg;
 }
 if (/\[\?NEXTPAGE(.)(.*?)\1(.*?)\?\]/)
 {
  $temp = ($current < $max) ? $2 : $3;
  $temp =~ s/\[!PAGENUM!\]/$current + 1/eg;
  s/\[\?NEXTPAGE(.*?)\?\]/$temp/seg;
 }

 # Any remaining [!PAGENUM!]s can be substituted happily now.
 s/\[!PAGENUM!\]/$current/eg;
}


# Post formatting -- this parses URLs and generally improves the text.
sub formatPost
{
 my $tmp = $_[0];

 # Insert spaces into very long words, to stop extreme page-widening.
 # The number in braces is the max word length, feel free to raise it.
 # ADDITION: I'm checking that the last 10 characters do not match any HTML tags/entities.
 #           Until I get a really good comment parser in here, it'll have to do.
 #           Otherwise, XXX<br /> can get changed to XXX<b r /> which bolds subsequent text :(.
 $tmp =~ s/(\S{70}[^\&<>]{10})/$1 /g;
 #$tmp =~ s/(\S{80})/$1 /g;
 # Repair breaks in HTML &xxxx; entities from the previous line - move spaces after them.
 #$tmp =~ s/(\&[^;]*) ([^;]*;)/$1$2 /g;

 # Lessons from Slashdot 101: .page .widening .posts .like .this.
 # So we forcefully insert spaces into those lines to provide break points, otherwise
 # malicious users can widen the browser window to thousands of pixels.
 $tmp =~ s/(\W+\S+\s+\W+\S+\s+\W+\S+\s+\W+\S+\s+)/$1&nbsp;/g;

 # Support 'BB' tag style links. Watch the last character of URIs.
 # Add http:// to [url]www.site.com which would normally link to current folder.
 $tmp =~ s/\]www\./\]http\:\/\/www\./ig;
 # Match [img]picture.gif and write IMG tag.
 $tmp =~ s/\[img\]([^"\s<>\]\[]+[^"\s<>\]\[\.\,\(\)])(\[\/img\])?/<img src="$1" \/>/ig;
 # Match [url]mysite.com and write A tag.
 $tmp =~ s/\[url\]([^"\s<>\]\[]+[^"\s<>\]\[\.\,\(\)])(\[\/url\])?/<a href="$1" rel="nofollow">$1<\/a>/ig;

 # TODO: Add more code tags here (quoting, etc).

 # Allow for indenting spaces at the start of each line and format lines nicely.
 # Feel free to raise the max indent level at the end (default 10)
 $tmp =~ s/<br \/>(\s+)/"<br \/>".("&nbsp;"x(length $1>10 ? 10 : length $1))/eg;

 return $tmp;
}










##### FORUM PAGE CONSTRUCTION FUNCTIONS #####


# Append output to global variable.
sub forum_out
{
 $out_str .= $_[0];
}


# Create the common page header - pass a title.
sub printHeader
{
 # Compose the regular header template.
 getTemplate('globalHeader');
 doSub("PAGETITLE", $_[0]);
 printTemplate();
}

# Similar for footer. Also, this is as good a place as any to output the content.
sub printFooter
{
 getTemplate('globalFooter');
 # For those who insist upon removing the credit link. Please, just leave this in there?
 # It is not much to ask, I hope.
 # And it keeps you on solid legal ground regarding the script License Agreement.
 if (!m/twinhelix/i)
 {
  forum_out(qq|<div style="text-align: center">
    <b>Powered by <a href="http://www.twinhelix.com">TwinHelix Forums</a></b>
   </div>
   |);
 }
 printTemplate();


 # OK, that's our complete page content generated. Now do headers and caching stuff!

 #forum_out("Pragma: no-cache\n");
 #forum_out("Cache-Control: private\n");

 # You may want to put "; charset=ABCD" after the content-type for other languages?
 # Or, if you're in a Mozilla-only or comparable environment, return application/xhtml+xml.
 print "Content-Type: text/html\n";

 # ETags are cool, since Last-Modified is an incredible pain with scripts like this.
 # As such, here's a quick'n'nasty hash generator that packs a string down to ASCII chars.
 # NOTE: I'm including the username in the hash, so users get a unique document each.
 my $etag_hash = '          ';
 my $etag = '';
 foreach my $content_chunk (split(/.{10}/, $user_name.$out_str)) { $etag_hash ^= $content_chunk }
 foreach my $etag_char (split(//, $etag_hash)) { $etag .= sprintf('%x', ord($etag_char)) }
 print "ETag: $etag\n";

 # Compare against a supplied ETag, if any.
 if ($ENV{'HTTP_IF_NONE_MATCH'} && $ENV{'HTTP_IF_NONE_MATCH'} =~ m/$etag/)
 {
  print "Status: 304 Not Modified\n";
  print "\n";
  return;
 }

 # Otherwise, if this is a new request, compressing and output the buffer as needed.
 if ($forum_prefs{'use_gzip'})
 {
  eval('$out_str = Compress::Zlib::memGzip($out_str)');
  print "Content-Encoding: gzip\n";
  print "Vary: Accept-Encoding,Cookie\n";
 }

 print "Content-Length: ".(length($out_str))."\n";
 print "\n";
 print $out_str;
}







# The first welcome screen, listing the available forums. Uses global @forum_data.
sub listForums
{
 # Start the HTML document from the template file.
 printHeader("Home");

 # Do common substitutions to the "page top" template, and print it.
 getTemplate('listForumsTop');
 printTemplate();

 # Some meaningless totals, to show how popular these forums are.
 my $totalTopics = 0;
 my $totalPosts = 0;

 # Loop through the forums list, write out rows for each forum.
 for (my $i = 0; $i <= $#forum_data; $i++)
 {
  # First of all, if this is a "section break" row, write that out and loop again.
  my $section = $forum_data[$i]{'section'};
  if ($section)
  {
   getTemplate('listForumsSection', \%{$forum_data[$i]});
   doSub("SECTIONTITLE", $section);
   printTemplate();
   next;
  }

  # Otherwise, read the 'data.txt' file for each forum listed.
  my @forum_stats = readFileData("$forum_prefs{'base_dir'}/$forum_data[$i]{'id'}/data.txt");

  # Add onto the total stats.
  $totalTopics += $forum_stats[0]{'topics'};
  $totalPosts += $forum_stats[0]{'posts'};

  # Make the appropriate substitutions from the forum data array and calculation above.
  getTemplate('listForumsRow', \%{$forum_data[$i]});
  my $lastauthor = $forum_stats[0]{'lastauthor'};
  if ($lastauthor =~ s/^<([^>]+)>/$1/)
  {
   doSub("LASTAUTHOR", "<a href=\"$forum_prefs{'path_name'}".
    ($forum_prefs{'path_mode'} ? "/_user/" : "?a=users&amp;u=").url_enc($lastauthor)."\">$lastauthor</a>");
  }
  else { doSub("LASTAUTHOR", $lastauthor) }
  my $lastTime = $forum_stats[0]{'lasttime'};
  s/\[!LASTTIMESECS!\]/$lastTime/eg;
  s/\[!LASTTIMEGMT!\]/gmtTimeFormat($lastTime)." GMT"/eg;
  s/\[\?HIGHLIGHTED(.)(.*?)\1(.*?)\?\]/(time - $lastTime < $forum_prefs{'highlight_forums'}*60*60) ?
   $2 : $3/seg;
  s/\[!TOPICS!\]/$forum_stats[0]{'topics'}/eg;
  s/\[!POSTS!\]/$forum_stats[0]{'posts'}/eg;
  printTemplate();
 }

 # Now write out the list closing tags and global footer, with substitutions.
 getTemplate('listForumsBottom');
 s/\[!TOTALTOPICS!\]/$totalTopics/eg;
 s/\[!TOTALPOSTS!\]/$totalPosts/eg;
 printTemplate();

 printFooter();
}








# Displays list of all topics in the current forum. No parameters, uses @topic_data.
sub listTopics
{
 # Check if the current user is entitled to do this.
 if (!$user_priv{'view'})
 {
  printHeader('Disallowed');
  getTemplate('userDisallowed', $forum_ref);
  doSub("ACTION", 'View topic listings');
  printTemplate();
  return printFooter();
 }

 # Do page headers -- display either the keywords or the forum name in the titlebar.
 printHeader($keywords ? "Searching \'$forum_ref->{'name'}\' for: $keywords" : $forum_ref->{'name'});

 # Read the index file for the selected topic directory if it hasn't been read
 # already by another function (e.g. posting messages).
 if ($#topic_data eq -1)
 {
  # Here's is a speed tweak: if we're displaying Page 1 of a forum, don't read the entire data
  # file, just the first page and pad out the array length appropriately.
  if (!$keywords && ($page == 1))
  {
   @topic_data = readFileData("$forum_prefs{'base_dir'}/$forum/index.txt",
    $forum_prefs{'topics_per_page'});
   my @forum_stats = readFileData("$forum_prefs{'base_dir'}/$forum/data.txt");
   $#topic_data = $forum_stats[0]{'topics'} - 1;
  }
  else
  {
   # Might as well just parse the entire file beyond that point.
   @topic_data = readFileData("$forum_prefs{'base_dir'}/$forum/index.txt");
  }
 }

 # Doesn't exist? Call updateIndex to create it; it will also populate @topic_data.
 if ($#topic_data eq -1)
 {
  updateIndex($forum);
 }

 # Now, if we have been passes search keywords, searchForum() will scan and reorder
 # the @topic_data array, so call that now.
 searchForum() if ($keywords);


 # Make sure we don't go too far, and set start/end topic variables.
 my $total_pages = int($#topic_data / $forum_prefs{'topics_per_page'}) + 1;

 if ($page < 1) { $page = 1 }
 if ($page > $total_pages) { $page = $total_pages }

 my $startVal = ($page - 1) * $forum_prefs{'topics_per_page'};
 my $endVal = $startVal + $forum_prefs{'topics_per_page'} - 1;
 if ($endVal > $#topic_data) { $endVal = $#topic_data }


 # The "topic list top" template, sub in values from arrays, and print.
 # We use the page numbers calculated above if the user has [!PAGE....!] markers.
 getTemplate('listTopicsTop', $forum_ref);
 if ($forum_prefs{'path_mode'})
 {
  s/\[!PAGEURL!\]/"$forum_prefs{'path_name'}\/$forum\/p=[!PAGENUM!]"/eg;
 }
 else { s/\[!PAGEURL!\]/"$forum_prefs{'path_name'}?f=$forum&amp;p=[!PAGENUM!]"/eg; }
 # Call the pagelist sub with the page info, remember it uses $_ as a source.
 formatPageLinks(1, $page, $total_pages);
 printTemplate();


 # Print an error if no topics have been found.
 if ($#topic_data eq -1)
 {
  getTemplate('listTopicsError', $forum_ref);
  printTemplate();
 }
 else
 {
  # Loop through array, displaying appropriate topics in table rows.
  for (my $i = $startVal; $i <= $endVal; $i++)
  {
   getTemplate('listTopicsRow', 0, 1);

   # Calculate the number of pages this topic will display when viewed.
   my $topic_pages = int(($topic_data[$i]{'posts'} - 1) / $forum_prefs{'posts_per_page'}) + 1;

   # Sub the appropriate values into the template and call "study" for speed.
   study;
   s/\[\?PINNED(.)(.*?)\1(.*?)\?\]/index($topic_data[$i]{'attr'}, 'P') >= 0 ? $2 : $3/eg;
   s/\[\?LOCKED(.)(.*?)\1(.*?)\?\]/index($topic_data[$i]{'attr'}, 'L') >= 0 ? $2 : $3/eg;
   s/\[\?HIGHLIGHTED(.)(.*?)\1(.*?)\?\]/($topic_data[$i]{'posts'}>=$forum_prefs{'highlight_topics'}) ?
    $2 : $3/seg;
   # Use our "safe" substitution routine for potentially tainted data.
   doSub("TOPICNAME", $topic_data[$i]{'topicname'});
   doSub("TOPICURL", $forum_prefs{'path_name'}. ($forum_prefs{'path_mode'} ?
 	"/$forum/" : "?f=$forum&amp;t="). $topic_data[$i]{'topicnum'});
   # Handle user accounts as links.
   my $origauthor = $topic_data[$i]{'origauthor'};
   if ($origauthor =~ s/^<([^>]+)>/$1/)
   {
    doSub("ORIGAUTHOR", "<a href=\"$forum_prefs{'path_name'}".
     ($forum_prefs{'path_mode'} ? "/_user/" : "?a=users&amp;u=").url_enc($origauthor)."\">$origauthor</a>");
   }
   else { doSub("ORIGAUTHOR", $origauthor) }
   my $lastauthor = $topic_data[$i]{'lastauthor'};
   if ($lastauthor =~ s/^<([^>]+)>/$1/)
   {
    doSub("LASTAUTHOR", "<a href=\"$forum_prefs{'path_name'}".
     ($forum_prefs{'path_mode'} ? "/_user/" : "?a=users&amp;u=").url_enc($lastauthor)."\">$lastauthor</a>");
   }
   else { doSub("LASTAUTHOR", $lastauthor) }
   s/\[!ORIGTIMESECS!\]/$topic_data[$i]{'origtime'}/eg;
   s/\[!ORIGTIMEGMT!\]/gmtTimeFormat($topic_data[$i]{'origtime'})." GMT"/eg;
   s/\[!LASTTIMESECS!\]/$topic_data[$i]{'lasttime'}/eg;
   s/\[!LASTTIMEGMT!\]/gmtTimeFormat($topic_data[$i]{'lasttime'})." GMT"/eg;
   s/\[!POSTS!\]/$topic_data[$i]{'posts'}/eg;
   # Same page link formatting routine as before, different link format.
   if ($forum_prefs{'path_mode'})
   {
    s/\[!PAGEURL!\]/"$forum_prefs{'path_name'}\/$forum\/$topic_data[$i]{'topicnum'}\/p=[!PAGENUM!]"/eg;
   }
   else
   {
    s/\[!PAGEURL!\]/"$forum_prefs{'path_name'}?f=$forum&amp;t=$topic_data[$i]{'topicnum'}&amp;".
	    "p=[!PAGENUM!]"/eg;
   }
   # We start listing from page 2 as the regular TOPICURL points to page 1.
   formatPageLinks(2, 1, $topic_pages);
   printTemplate();
  }
 }

 # Display the bottom of topic list template, again with substitutions.

 getTemplate('listTopicsBottom', $forum_ref);
 if ($forum_prefs{'path_mode'})
 {
  s/\[!PAGEURL!\]/"$forum_prefs{'path_name'}\/$forum\/p=[!PAGENUM!]"/eg;
 }
 else { s/\[!PAGEURL!\]/"$forum_prefs{'path_name'}?f=$forum&amp;p=[!PAGENUM!]"/eg; }
 formatPageLinks(1, $page, $total_pages);
 printTemplate();

 printFooter();
}




# This function is called by listTopics() above when keywords have been posted.
# It scans and reconstructs the @topic_data array to reflect the search results.
sub searchForum
{
 # The 'keywords' global variables has alreayd been picked up from the post data.
 # Split it into an array, throw error if no keywords found, and limit to 10 keywords max.
 my @search_terms = split(/\s+/, $keywords);
 if ($#search_terms == -1) { @topic_data = (); return }
 if ($#search_terms > 10) { $#search_terms = 10 }

 # An array, which will store the numbers of searched topics in subarrays.
 # Index in this array = final score of topic, so it's easy to output sorted lists.
 my @topic_rank = ();

 # Now, loop through @topic_data and compare search_terms. Count message bodies searched.
 my $messages_searched = 0;

 for (my $t = 0; $t <= $#topic_data; $t++)
 {
  my $score = 0;
  $messages_searched++;

  # Retrieve the posts from this topic file, if we're supposed to search it.
  my @msg_body = ();
  if ($messages_searched <= $forum_prefs{'max_search_messages'})
  {
   @msg_body = readFileData("$forum_prefs{'base_dir'}/$forum/$topic_data[$t]{'topicnum'}.txt");
  }

  # Increment score for each match of each keyword against the topic.
  foreach my $kw (@search_terms)
  {
   # Only allow words >= 3 characters in length.
   next if (length $kw < 3);

   # Always search topic titles, they're cached and pretty quick.
   # Only match search_terms at the start of topic/word (allow plurals etc. after match).
   $score++ while ($topic_data[$t]{'topicname'} =~ /(^| )\Q$kw\E/isg);

   # If we've got the messages within the topic, search them.
   if ($#msg_body > -1)
   {
    for (my $i = 1; $i <= $#msg_body; $i++)
   	{
	    $score++ while ($msg_body[$i]{'message'} =~ /(^| )\Q$kw\E/isg);
     # I'm gonna search author names too. Because I can.
     $score++ if ($msg_body[$i]{'author'} =~ /\Q$kw\E/i);
	   }
   }
  }

  # Now, if this has a positive score, assign it to the rank array.
  if ($score)
  {
   # Each entry in the rank array itself must be an array... can have multiples!
   unless ($#{$topic_rank[$score]} > -1) { $topic_rank[$score] = [] }
   # Add the index of this topic within @topic_data to this rank position.
   push @{$topic_rank[$score]}, $t;
  }
 }

 # What will become our new topic_data array.
 my @search_results = ();
 # Loop backwards through @topic_rank and add valid results to @search_results.
 RANKLOOP: for (my $r = $#topic_rank; $r > 0; $r--)
 {
  next if (!$topic_rank[$r]);
  # Each entry in the rank array is an array itself, so loop through that.
  foreach my $topicnum (@{$topic_rank[$r]})
  {
   # Until I come up with a really clever way of paginating search results, this'll do.
   last RANKLOOP if ($#search_results >= ($forum_prefs{'topics_per_page'} - 1));
   $search_results[$#search_results + 1] = $topic_data[$topicnum];
  }
 }

 # Finally, overwrite @topic_data with our new array, and return to listTopics().
 @topic_data = @search_results;
}











# Displays 'New Topic' page, which submits the form from printPostForm() to postMessage().
sub newTopic
{
 # Check if the current user is entitled to do this.
 if (!$user_priv{'post'})
 {
  printHeader('Disallowed');
  getTemplate('userDisallowed', $forum_ref);
  doSub("ACTION", 'Create new topics');
  printTemplate();
  return printFooter();
 }

 printHeader($forum_ref->{'name'});

 # Retrieve and format the "New Topic" template.
 getTemplate('newTopicTop', $forum_ref);
 printTemplate();

 # Just in case, remove the existing $topic setting, and display the rest of the form.
 $topic = 0;
 printPostForm();

 printFooter();
}



# Displays the 'new post' form, creates under $topic if that exists, or else a new topic.
# You can pass default values for the fields if you wish -- topic name, author, message.
sub printPostForm
{
 my ($topicname, $author, $message, $num) = @_;

 # Set the proper action type for the hidden fields, whether this is a new topic or reply.
 # Include an antispam fake email and a timestamp (TODO: secure nonce).
 my $timestamp = crypt(reverse(time()), "TS");
 my $hiddenfields = qq|
  <label style="display: none">Please ignore: <input type="text" name="email" id="email" /></label>
  <input type="hidden" name="ts" value="$timestamp" />
  <input name="a" type="hidden" value="postmessage" />
  <input name="f" type="hidden" value="$forum" />|;

 # Pass our topic number if available, and set page=9999 (so we always go to the newest page).
 $hiddenfields .= qq|\n  <input name="t" type="hidden" value="$topic" />| if ($topic);
 # The individual post number, for editing posts.
 $hiddenfields .= qq|\n  <input name="n" type="hidden" value="$num" />| if ($num);

 # Retrieve and format template.
 getTemplate('postForm', $forum_ref);
 if ($forum_prefs{'path_mode'})
 {
  doSub("POSTURL", $forum_prefs{'path_name'}.($topic ? "/$forum/$topic" : ""));
 }
 else
 {
  doSub("POSTURL", $forum_prefs{'path_name'}.($topic ? "?f=$forum&amp;t=$topic" : ""));
 }
 s/\[!POSTFIELDS!\]/$hiddenfields/eg;
 s/\[\?ISNEWTOPIC(.)(.*?)\1(.*?)\?\]/!$topic ? $2 : $3/seg;
 s/\[\?ISNEWPOST(.)(.*?)\1(.*?)\?\]/$topic && !$num ? $2 : $3/seg;
 # If !TOPICNAME! hasn't been already processed by the getTemplate() function...
 doSub("TOPICNAME", $topicname);
 # Return the message to the user in case of posting error.
 doSub("AUTHOR", $author);
 doSub("MESSAGE", $message);
 printTemplate();
}











# Displays all messages in the current topic from the URL query.
sub displayTopic
{
 # Check entitlements.
 if (!$user_priv{'view'})
 {
  printHeader('Disallowed');
  getTemplate('userDisallowed', $forum_ref);
  doSub("ACTION", 'View posts in this topic');
  printTemplate();
  return printFooter();
 }

 # Read information on selected topic, if another function hasn't already read it.
 if ($#post_data == -1)
 {
  @post_data = readFileData("$forum_prefs{'base_dir'}/$forum/$topic.txt");
 }

 if ($#post_data eq -1)
 {
  # Print the topic error template if we can't find the datafile.
  printHeader("Topic Error");

  getTemplate('displayTopicError', $forum_ref);
  printTemplate();

  # Update the forum index, to remove any broken links.
  updateIndex($forum);
 }
 else
 {
  # We have a valid topic, so let's process and display it.
  # Sub topic name into header & print it.
  printHeader($post_data[0]{'name'});

  # Sort out what page we're on, and the numbers we'll display in the loop.

  my $total_pages = int(($#post_data - 1) / $forum_prefs{'posts_per_page'}) + 1;

  if ($page < 1) { $page = 1 }
  if ($page > $total_pages) { $page = $total_pages }

  my $startVal = ($page - 1) * $forum_prefs{'posts_per_page'} + 1;
  my $endVal = $startVal + $forum_prefs{'posts_per_page'} - 1;
  if ($endVal > $#post_data) { $endVal = $#post_data }


  # Retrieve the top of the page template, sub in appropriate values.
  getTemplate('displayTopicTop', $forum_ref);
  # The getTemplate function will have substituted TOPICNAME and TOPICURL for us.
  if ($forum_prefs{'path_mode'})
  {
   s/\[!PAGEURL!\]/"$forum_prefs{'path_name'}\/$forum\/$topic\/p=[!PAGENUM!]"/eg;
  }
  else
  {
   s/\[!PAGEURL!\]/"$forum_prefs{'path_name'}?f=$forum&amp;t=$topic&amp;p=[!PAGENUM!]"/eg;
  }
  formatPageLinks(1, $page, $total_pages);
  printTemplate();


  # Loop through all posts in the current topic info array.
  for (my $i = $startVal; $i <= $endVal; $i++)
  {
   # Format the IP address of the user -- hide the last part of the subnet for non-admins.
   my $ip_addr = $post_data[$i]{'ip'};
   if (!$user_num || !$user_priv{'anyedit'}) 
   {
    $ip_addr =~ s/\W*(\w*)(\.|:)(\w*)(\.|:).*/$1$2$3$4---/;
   }
   else
   {
    # Admins get an IP lookup. Please don't abuse this service...
	   $ip_addr = "<a href=\"http://ip-lookup.net/index.php?ip=$ip_addr\">$ip_addr</a>";
   }

   # Call the "formatPost" function to handle URLs and so forth.
   my $message = formatPost($post_data[$i]{'message'});

   # Now, write out our formatted row in the appropriate template.
   getTemplate('displayTopicRow', 0, 0);
   doSub('NUM', $i);

   # Detect posts by forum members, stored in the for <username>.
   my $author = $post_data[$i]{'author'};
   if ($author =~ s/^<([^>]+)>/$1/)
   {
    # Read in the user data for this account if required.
   	my $acc_num = $user_index{lc $author};
    if (!$user_data[$acc_num])
    {
     my @tmp = readFileData("$forum_prefs{'user_dir'}/$acc_num.txt");
     $user_data[$acc_num] = \%{$tmp[0]};
    }
    # And substitute it in.
    doSub("AUTHOR", "<a href=\"$forum_prefs{'path_name'}".
     ($forum_prefs{'path_mode'} ? "/_user/" : "?a=users&amp;u=").
     url_enc($author)."\">$author</a>");
    s/\[\?ISUSER(.)(.*?)\1(.*?)\?\]/$2/seg;
    doSub("POSTS", $user_data[$acc_num]{'posts'});
    doSub("JOINED", gmtTimeFormat($user_data[$acc_num]{'joined'}, 1));
    doSub("LOCATION", $user_data[$acc_num]{'location'});
    s/\[\?USERSIG(.)(.*?)\1(.*?)\?\]/$user_data[$acc_num]{'sig'} ? $2 : $3/seg;
    doSub("SIG", "$user_data[$acc_num]{'sig'}");
    doSub("GROUP", $user_groups{$user_data[$acc_num]{'group'}}{'name'});
   }
   else
   {
    doSub("AUTHOR", $author);
    s/\[\?ISUSER(.)(.*?)\1(.*?)\?\]/$3/seg;
    s/\[\?USERSIG(.)(.*?)\1(.*?)\?\]//seg;
    doSub("GROUP", $user_groups{'guests'}{'name'});
    doSub("SIG", '');
   }

   # Create links to this post to edit it, if required. Use post time if the 'num' is unavailable.
   my $is_editable = ((($user_priv{'ownedit'} && $post_data[$i]{'author'} eq "<$user_name>") ||
    $user_priv{'anyedit'}) && ($user_priv{'topicset'} || index($post_data[0]{'attr'}, 'L') == -1)) ?
	   1 : 0;
   s/\[\?ISEDITABLE(.)(.*?)\1(.*?)\?\]/$is_editable ? $2 : $3/seg;
   if ($is_editable)
   {
    my $post_id = $post_data[$i]{'num'} || $post_data[$i]{'time'};
    doSub("EDITURL", "$forum_prefs{'path_name'}". ($forum_prefs{'path_mode'} ?
     "/$forum/$topic/edit=" : "?f=$forum&amp;t=$topic&amp;a=edit&amp;n=").$post_id);
   }

   # The rest of these are rather boring.
   doSub("IPADDR", $ip_addr);
   doSub("TIMESECS", $post_data[$i]{'time'});
   doSub("TIMEGMT", gmtTimeFormat($post_data[$i]{'time'})." GMT");
   doSub("MESSAGE", $message);
   printTemplate();
  }


  # Now format the bottom -- same substitutions as the top.
  getTemplate('displayTopicBottom', $forum_ref);
  if ($forum_prefs{'path_mode'})
  {
   s/\[!PAGEURL!\]/"$forum_prefs{'path_name'}\/$forum\/$topic\/p=[!PAGENUM!]"/eg;
  }
  else
  {
   s/\[!PAGEURL!\]/"$forum_prefs{'path_name'}?f=$forum&amp;t=$topic&amp;p=[!PAGENUM!]"/eg;
  }
  formatPageLinks(1, $page, $total_pages);
  printTemplate();


  # Show this on the last page of the post listing, assuming the user's entitled & it's not locked.
  # I'm allowing admin users to post to locked topics, as well.
 if ($page == $total_pages)
 {
  if ($user_priv{'post'} &&
      (index($post_data[0]{'attr'}, 'L') == -1 || $user_priv{'topicset'}) )
  {
   printPostForm();
  }
  else
  {
   getTemplate('displayTopicNoPost', $forum_ref);
   printTemplate();
  }
 }

  # The topic settings form.
  printTopicSettings() if ($user_priv{'topicset'});
 }

 # The footer is always printed, whether the topic has been found or not.
 printFooter();
}









# The topic settings form functions to print and handle input from said form.
# The form display is called from within "displayTopic" above.

sub printTopicSettings
{
 # It's four lines, do you really want me to comment this? ;).
 getTemplate('topicSettings', $forum_ref);
 doSub('HIDDENFIELDS',
 qq| <input type="hidden" name="a" value="topicsettings" />
 <input type="hidden" name="f" value="$forum" />
 <input type="hidden" name="t" value="$topic" />|);

 s/\[\?LOCKED(.)(.*?)\1(.*?)\?\]/index($post_data[0]{'attr'}, 'L') >= 0 ? $2 : $3/seg;
 s/\[\?PINNED(.)(.*?)\1(.*?)\?\]/index($post_data[0]{'attr'}, 'P') >= 0 ? $2 : $3/seg;
 printTemplate();
}

sub processTopicSettings
{
 # Handles input from the above form, makes appropriate changes to the topic data.

 # Check privileges
 displayTopic() if (!$user_priv{'topicset'});

 # In with the data for this topic.
 @post_data = readFileData("$forum_prefs{'base_dir'}/$forum/$topic.txt");

 # Acquire the CGI parameters, and fix the topic up.
 if ($query->param('delete'))
 {
  unlink "$forum_prefs{'base_dir'}/$forum/$topic.txt";
  updateIndex($forum);
  return listTopics();
 }
 else
 {
  $post_data[0]{'name'} = ($query->param('topicname') =~ m/^(.{1,60})/) ? trimHTML($1) : "";
  $post_data[0]{'attr'} = ($query->param('locked')?'L':'').($query->param('pinned')?'P':'');
 }

 # Now write, index and display the updated topic.
 writeFileData("$forum_prefs{'base_dir'}/$forum/$topic.txt", \@post_data);
 updateIndex($forum, $topic);
 displayTopic();
}






# Handles editing posts, called from the links generated above.
sub editPost
{
 # This is a basic check, postMessage() is more thorough.
 if (!$user_priv{'ownedit'} && !$user_priv{'anyedit'})
 {
  printHeader('Disallowed');
  getTemplate('userDisallowed', $forum_ref);
  doSub("ACTION", 'Edit posts in this topic');
  printTemplate();
  return printFooter();
 }

 # Retrieve the post ID from the URL query or form submittal.
 my $post_num = $query->param('n') || ($path_info[3] =~ m/^edit=(.*)/ ? $1 : 0);

 # In case of silly pranksters, just revert to displaying the information.
 if (!$post_num) { displayTopic() }

 # Read information on selected topic, and loop through it to find the right post.
 if ($#post_data == -1)
 {
  @post_data = readFileData("$forum_prefs{'base_dir'}/$forum/$topic.txt");
 }
 my $post_ref = 0;
 foreach my $post (@post_data)
 {
  $post_ref = $post if ($post->{'num'} == $post_num || $post->{'time'} == $post_num);
 }
 # In case of already deleted posts, mistyped URLs, or 1337 haxors.
 if (!$post_ref || (!$user_priv{'topicset'} && index($post_data[0]{'attr'}, 'L') >= 0))
 {
  displayTopic();
 }

 printHeader('Editing Post');
 getTemplate('editPost', $forum_ref);
 printTemplate();
 my $message = $post_ref->{'message'};
 $message =~ s/<br \/>/\n/g;
 printPostForm('', $post_ref->{'author'}, $message, $post_num);
 printFooter();
}






# Handles creating/updating messages and topics -- checks CGI parameters and updates files.
sub postMessage
{
 # Get $topicname and $message from URL, converting HTML to &entities;
 my $topicname = ($query->param('topicname') =~ m/^(.{1,60})/) ? trimHTML($1) : "";
 # Pass '1' as parameter to trimHTML, so we allow linebreaks.
 my $message = trimHTML($query->param('message'), 1);
 my $edit_num = 0;
 # Records errors in the submitted information.
 my $post_error = "";
 my $spam_error = 0;

 # Check user posting/editing entitlements. I'm displaying errors using this sub's own template.
 if ($query->param('n') && !$user_priv{'ownedit'} && !$user_priv{'anyedit'})
 {
  $post_error .= "<div>Sorry, you are not allowed to edit posts with that account.</div>\n";
 }
 elsif (!$user_priv{'post'})
 {
  $post_error .= "<div>Sorry, you are not allowed to post messages with that account.</div>\n";
 }

 # The fake "email" field is only filled in by spambots ;).
 if ($query->param('email'))
 {
  $spam_error = 1;
 }

 # Check the submitted timestamp isn't within the last few seconds (non-admins).
 if (!$query->param('ts'))
 {
  $spam_error = 1;
 }
 elsif (!$user_priv{'anyedit'})
 {
  for (my $i = 0; $i < 6; $i++)
  {
   my $timestamp = crypt(reverse(time() - $i), "TS");
   if ($query->param('ts') eq $timestamp)
   {
    $post_error .= "<div>Slow down - you are posting too fast!</div>\n";
    last;
   }
  }
 }

 # Logged in users get their name recorded in angle brackets for now.
 # For anonymous posters, get the author and remove illegal chars.
 my ($author, $tmp_auth);
 if ($user_name) { $tmp_auth = $author = "<$user_name>" }
 else
 {
  $author = ($query->param('author') =~ m/^(.{1,25})/) ? $1 : "";
  $tmp_auth = $author;
  $tmp_auth =~ s/[^\Q$forum_prefs{'name_chars'}\E]//g if $forum_prefs{'name_chars'};
  $tmp_auth =~ s/(\s{2,})/\ /g;
  $tmp_auth =~ s/^\s*(.*?)\s*$/$1/g;
  # Pass both through trimHTML() in case any nasties slip by.
  $author = trimHTML($author);
  $tmp_auth = trimHTML($tmp_auth);
 }

 # Loop through topic folder, searching for $topic and also highest topic number.
 # The highest topic number found is used to create a new topic file if needed.
 my @dirlines = ();
 my $topic_found = 0;
 my $new_topic_num = 1;
 if (opendir (TOPICSDIR,"$forum_prefs{'base_dir'}/$forum"))
 {
  @dirlines = readdir(TOPICSDIR);
  closedir(TOPICSDIR) or die "Could not close topics dir: $forum, $!";
 }

 my $topicFile;
 foreach $topicFile (@dirlines)
 {
  if ($topicFile eq "$topic.txt") { $topic_found = 1 }
  if ($topicFile =~ m/^(\d*)/)
  {
   if ($new_topic_num <= $1) { $new_topic_num = $1 + 1 }
  }
 }


 # Identify if we're legally editing an individual post within this topic.
 my $edit_num = 0;
 if ($topic_found)
 {
  # Read in the topic data.
  @post_data = readFileData("$forum_prefs{'base_dir'}/$forum/$topic.txt");
  # The number to edit from the query. However, we must check it...
  my $num = $query->param('n');
  # Scan through that, if we match a post, record its array index in $edit_num.
  # I'm looping backwards for speed... I assume most editable posts will be recent.
  for (my $i = $#post_data; $i > 0; $i--)
  {
   $edit_num = $i if ($num && ($post_data[$i]{'num'} == $num || $post_data[$i]{'time'} == $num));
  }
  # Check we're entitled to edit it...
  if ($edit_num && ($post_data[$edit_num]{'author'} ne "<$user_name>") && !$user_priv{'anyedit'})
  {
   $post_error .= "<div>Sorry, you aren't allowed to edit that post.</div>\n";
  }
 }



 # Now check for improperly filled out forms etc. Most of these we skip for editing existing posts.
 if ($ENV{'REQUEST_METHOD'} ne 'POST')
 {
  $post_error .= "<div>Sorry, posts must be made using form submittals.</div>\n";
 }
 if (index($post_data[0]{'attr'}, 'L') >= 0 && !$user_priv{'topicset'})
 {
  $post_error .= "<div>This topic is locked, and posts cannot be submitted.</div>\n";
 }
 if (!$author && !$edit_num)
 {
  $post_error .= "<div>Please enter a name, nickname or similar before posting.</div>\n";
 }
 if (!$message && !$edit_num)
 {
  $post_error .= "<div>Please enter a message before posting.</div>\n";
 }
 if (length($query->param('message')) > $forum_prefs{'max_post_len'})
 {
  $post_error .= "<div>Sorry, the maximum post length is $forum_prefs{'max_post_len'} characters.\n
   Please trim it down a little; your post is currently ".
   (length $query->param('message'))." bytes long.</div>\n";
 }
 if ((length($message) > 100))
 {
  my $links = 0;
  $links++ while ($message =~ /http\:/g);
  if ($links > (length($message) / 200))
  {
   $post_error .= "<div>Please don't spam - too many links.</div>\n";
  }
 }
 if ($message =~ /\[url=/)
 {
  $post_error .= "<div>Please don't spam - use the correct link syntax.</div>";
 }

 if (!$topicname && !$topic_found)
 {
  $post_error .= "<div>To create a new topic, you must give it a name to display.</div>\n";
 }
 if ($topicname && $topic_found)
 {
  $post_error .= "<div>You can't overwrite existing topics with your own new topics.</div>\n";
 }
 if (!$topic_found)
 {
  # For new topics, check for duplicates (matching origauthor and topicname).
  # Read the index data for this forum and scan through the last 20 posts.
  @topic_data = readFileData("$forum_prefs{'base_dir'}/$forum/index.txt");
  my $duplicatePost = 0;
  for (my $t = 0; $t <= ($#topic_data > 20 ? 20 : $#topic_data); $t++)
  {
   if (($author eq $topic_data[$t]{'origauthor'}) && ($topicname eq $topic_data[$t]{'topicname'}))
   {
    my $err_uri = $forum_prefs{'path_name'}.($forum_prefs{'path_mode'} ? "?f=" : "/")."$forum";
    $post_error .= "<div>This appears to be a duplicate topic -- another one is already registered
     with that title under your name. Make sure you aren't refreshing/reposting pages, and check
     the <a href=\"$err_uri\">topic listing</a> for an earlier post.</div>\n";
   }
  }
 }
 elsif (!$edit_num)
 {
  # For new replies to existing topics, check if this is a dupe reply (post then refresh bug)...
  if (($author eq $post_data[$#post_data]{'author'}) && ($message eq $post_data[$#post_data]{'message'}))
  {
   my $last_page = int($#post_data / $forum_prefs{'posts_per_page'}) + 1;
   my $err_uri = $forum_prefs{'path_name'};
   if ($forum_prefs{'path_mode'}) { $err_uri .= "/$forum/$topic/p=$last_page" }
   else { $err_uri .= "?f=$forum&amp;t=$topic&amp;p=$last_page" }
   $post_error .= "<div>This appears to be a duplicate post. Have a look at the existing ".
    "<a href=\"$err_uri\">posts in this topic</a> and see if your post is there already.</div>\n";
  }
 }

 # Got an error after all of that? Display message and post form, then return.
 if ($post_error)
 {
  printHeader("Post Error");
  getTemplate('postError', $forum_ref);
  doSub("ERRORMESSAGE", $post_error);
  printTemplate();
  # Reformat the message for client-side display.
  $message =~ s/<br \/>/\n/g;
  printPostForm($topicname, $author, $message, $edit_num);
  printFooter();
  return;
 }

 # Folder doesn't exist? Make it.
 if ($#dirlines == -1) { mkdir("$forum_prefs{'base_dir'}/$forum", 0755) }

 # Right, let's actually post this message at last.
 # Ensure we have the right number, and create a new topic descriptor if needed.
 if (!$topic_found)
 {
  $topic = $new_topic_num;
  $post_data[0] = { 'name' => $topicname, 'attr' => '' };
 }

 # Now: are we editing an existing post or not?
 if ($edit_num)
 {
  # Existing. Check if we're changing a post or deleting it entirely.
  if ($message)
  {
   $post_data[$edit_num]{'message'} = $message;
  }
  else
  {
   # Deleting. Check if from user, if so decrement post counter
   my $edit_user_name = $post_data[$edit_num]{'author'};
   if ($edit_user_name =~ s/^<(.*)>$/$1/)
   {
    my $edit_user_num = $user_index{lc $edit_user_name};
    if ($edit_user_num)
    {
     # Retrieve the info for this username from the appropriate numbered filename.
     @tmp = readFileData("$forum_prefs{'user_dir'}/$edit_user_num.txt");
     $user_data[$edit_user_num] = \%{$tmp[0]};
     $user_data[$edit_user_num]{'posts'}--;
     writeFileData("$forum_prefs{'user_dir'}/$edit_user_num.txt", [$user_data[$edit_user_num]]);
    }
   }
   # Deleting: remove the array entry entirely, and delete the file if that was the last post.
   splice (@post_data, $edit_num, 1);
   if ($#post_data < 1) { unlink "$forum_prefs{'base_dir'}/$forum/$topic.txt" }
  }
 }
 else
 {
  # Regular new topics or replies: append to the end of the array.
  $post_data[$#post_data + 1] = {
   'time' => time(),
   'ip' => $ENV{'REMOTE_ADDR'},
   'author' => $author,
   'message' => $message,
   # Create a new post number to uniquely identify this post.
   'num', => $post_data[$#post_data]{'num'} + 1
  };
  # And increment the user's post counter, assuming legit post :).
  if ($user_num && !$spam_error)
  {
   $user_data[$user_num]{'posts'}++;
   writeFileData("$forum_prefs{'user_dir'}/$user_num.txt", [$user_data[$user_num]]);
  }
 }


 # Spammers don't actually get their data written to disk ;).
 if (!$spam_error)
 {
  # Write out our new data, over the existing topic or creating a new one.
  writeFileData("$forum_prefs{'base_dir'}/$forum/$topic.txt", \@post_data) if ($#post_data > 0);
  # Update the forum index for this modified topic.
  updateIndex($forum, $topic);
 }
 # Finally, display the topic(s), and pick a hopefully-correct page (9999 = end of topic).
 if ($#post_data < 1) { listTopics() }
 else
 {
  $page = $edit_num ? int($edit_num/($forum_prefs{'posts_per_page'}+1)) : 9999;
  displayTopic();
 }
}












# USER MANAGEMENT FUNCTIONS

sub getCookiePathExp
{
 # This is just a handy function to return a correctly-formatted cookie path/date string.
 my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday) = gmtime($_[0]);
 $sec = sprintf("%02d", $sec);
 $min = sprintf("%02d", $min);
 $hour = sprintf("%02d", $hour);
 $mon = ('Jan','Feb','Mar','Apr','May','Jun','Jul','Aug','Sep','Oct','Nov','Dec')[$mon];
 $year = sprintf("%02d", $year % 100);
 $wday = ('Sunday','Monday','Tuesday','Wednesday','Thursday','Friday','Saturday')[$wday];
 return "path=$forum_prefs{'cookie_path'}; expires=$wday, $mday-$mon-$year $hour:$min:$sec GMT\n";
}


sub userLogin
{
 # Retrieve the user info from the post information, and user number from local data.
 my $p_user = $query->param('user');
 my $p_pass = $query->param('pass');
 my $unum = $user_index{lc $p_user} || 0;

 my $login_error = 0;

 # Have we heard of this user before? user_index is a hash of usernames.
 if ($unum)
 {
  # Read the user data from disk if needed, place in the user_data hash.
  if (!$user_data[$unum])
  {
   my @tmp = readFileData("$forum_prefs{'user_dir'}/$unum.txt");
   $user_data[$unum] = \%{$tmp[0]};
  }
  # Check the provided password against the stored password.
  $login_error = 1 unless ($user_data[$unum]{'pass'} eq crypt($p_pass, $user_data[$unum]{'pass'}));
 }
 else { $login_error = 1 }

 # Now, either display an error, or store the new login info
 if ($login_error)
 {
  printHeader('Login Error');
  getTemplate('userLoginError');
  printTemplate();
  printFooter();
 }
 else
 {
  # Remember the remote IP address for this login.
  $user_data[$unum]{'ip'} = $ENV{'REMOTE_ADDR'};
  # Generate a new, random session ID.
  my $new_session_id = '';
  foreach (1..20) { $new_session_id .= ('A'..'Z','a'..'z',0..9)[rand 62] }
  $user_data[$unum]{'sessionid'} = $new_session_id;
  # Record the current login time.
  $user_data[$unum]{'lastlogin'} = time;
  # And store all this on disk, to validate the user on future requests.
  # Pass the sub an array reference containing a hash reference.
  writeFileData("$forum_prefs{'user_dir'}/$unum.txt", [$user_data[$unum]]);

  # Record this user as logged-in, by setting the global variables.
  # $user_name is stored in URL-encoded form.
  $user_name = $user_data[$unum]{'name'};
  $user_num = $unum;
  # Set the appropriate privileges.
  %user_priv = %{$user_groups{$user_num?$user_data[$user_num]{'group'}:'guests'}};

  # Now, just print() the user's cookie, before main output commences.
  my $pathexp = getCookiePathExp(time + $forum_prefs{'session_length'}*60*60);
  print "Set-Cookie: user=" . url_enc($user_name) . ".$new_session_id; $pathexp";

  # And now, we display a page (hopefully) related to the one from which the user came.
  # If we came from the account editing function, return to there.
  if ($act eq 'account' || $path_info[2] eq '_account') { return }
  elsif ($topic) { displayTopic() }
  elsif ($forum) { listTopics() }
  else { listForums() }
 }
}


sub userLogout
{
 # This is the easy one. Set cookies in the past to delete.
 my $pathexp = getCookiePathExp(time - 100000);
 print "Set-Cookie: user=; $pathexp";
 # Clear out global user vars, return to the front page.
 $user_name = '';
 $user_num = 0;
 listForums();
}


sub userAccountView
{
 # Check entitlements.
 if (!$user_priv{'accview'})
 {
  printHeader('Disallowed');
  getTemplate('userDisallowed');
  doSub("ACTION", 'View user accounts');
  printTemplate();
  return printFooter();
 }

 # This just display information on the selected username in the URL.
 my $view_user = url_dec($query->param('u') || $path_info[2] || '');


 if ($view_user)
 {
  my $view_num = $user_index{lc $view_user};
  if (!$view_num)
  {
   # TODO: Implement a "user list" page here.
   printHeader('Error');
   forum_out('<p>Error: User not found.</p>');
   printFooter();
   return;
  }

  # Read in the account data for the selected user if it hasn't been already.
  if (!$user_data[$view_num])
  {
   my @tmp = readFileData("$forum_prefs{'user_dir'}/$view_num.txt");
   $user_data[$view_num] = \%{$tmp[0]};
  }
  # And print out the info.
  printHeader($user_data[$view_num]{'name'});
  getTemplate('userAccountView');
  s/\[\?ACCEDIT(.)(.*?)\1(.*?)\?\]/$user_priv{'accedit'} ? $2 : $3/seg;
  doSub('ACCOUNTUSER', $user_data[$view_num]{'name'});
  doSub('THISACCOUNTURL', $forum_prefs{'path_name'}.
   ($forum_prefs{'path_mode'} ? "/_account/" : "?a=account&amp;u=").url_enc($view_user));
  doSub("JOINED", gmtTimeFormat($user_data[$view_num]{'joined'}, 1));
  doSub('LOCATION', $user_data[$view_num]{'location'});
  doSub('BIO', formatPost($user_data[$view_num]{'bio'}));
  doSub('SIG', formatPost($user_data[$view_num]{'sig'}));
  doSub('POSTS', $user_data[$view_num]{'posts'}||'None');
  printTemplate();
  printFooter();
 }
 else
 {
  # TODO: Implement list of active user accounts here. In the meantime...
  listForums();
 }
}


sub userAccountEdit
{
 # This sub has several functions. It displays a form suitable for creating or editing an account
 # (depending on whether a user is currently logged in or not) and processes input from that form.

 # First of all, retrieve the posted info from the form POST if available.
 # We also accept usernames from the path info; they're escaped in the URL.
 my $p_user = (($query->param('user')||$query->param('u')||$path_info[2]||'') =~
  m/(.{1,25})/ ? url_dec($1) : '');
 my $p_pass = (($query->param('pass')||'') =~ m/(.{1,25})/ ? $1 : '');
 my $p_pass2 = (($query->param('pass2')||'') =~ m/(.{1,25})/ ? $1 : '');
 my $location = trimHTML($query->param('location')||'');
 my $bio = trimHTML($query->param('bio')||'', 1);
 my $sig = trimHTML($query->param('sig')||'', 1);
 # This is the variable we use to record any errors in the account creation/update process.
 my $acc_error = '';
 # Silent fail for spambots
 my $spam_error = 0;

 # Generate an antispam fake email and a timestamp (TODO: secure nonce).
 my $timestamp = crypt(reverse(time()), "TS");
 my $hiddenfields = qq|
 <div style="display: none">
  <label for="email">Email:</label>
  <input type="text" name="email" id="email" value="" />
 </div>
 <input type="hidden" name="ts" value="$timestamp" />|;

 # Remove some spaces/dodgy characters from the username, and store a temporary new copy.
 my $tmp_user = $p_user;
 $tmp_user =~ s/[^\Q$forum_prefs{'name_chars'}\E]//g if $forum_prefs{'name_chars'};
 $tmp_user =~ s/(\s{2,})/\ /g;
 $tmp_user =~ s/^\s*(.*?)\s*$/$1/g;
 # Pass both through trimHTML() in case any nasties slip by.
 $p_user = trimHTML($p_user);
 $tmp_user = trimHTML($tmp_user);

 # Work out what we're doing. Admins are allowed to edit any account.
 my $is_new_acc = !$user_name ? 1 : 0;
 my $is_own_acc = ($user_name && ($user_name eq $p_user)) ? 1 : 0;
 # Variables to hold the name/number of the target account.
 my $acc_name = $is_own_acc ? $user_name : $p_user;
 my $acc_num = $is_own_acc ? $user_num : ($is_new_acc ? 0 : $user_index{lc $p_user});

 # For admins editing other accounts, ensure the other account data is loaded.
 if ($user_priv{'accedit'} && !$user_data[$acc_num])
 {
  my @tmp = readFileData("$forum_prefs{'user_dir'}/$acc_num.txt");
  $user_data[$acc_num] = \%{$tmp[0]};
 }
 elsif (!$is_new_acc && !$is_own_acc)
 {
  # Other situations: return an info page if creation/modification is not allowed.
  return userAccountView();
 }


 # Do a dummy run on the post info, decide if it's all fine now, before printing anything.

 if ($ENV{'REQUEST_METHOD'} eq 'POST')
 {
  # If we're creating a new account (or no username given), run some basic tests first.
  if ($is_new_acc || !$p_user)
  {
   $acc_error .= "<div>Please specify a user name.</div>\n" if (!$p_user);

   # Check if any characters had to be removed.
   if ($p_user ne $tmp_user)
   {
    $acc_error .= "<div>Your username contained invalid characters that have been removed.</div>\n";
   }
   # Fix it up for the form output.
   $p_user = $tmp_user;

   # Only one user per username, please.
   $acc_error .= "<div>Sorry, that username is already taken.</div>\n" if ($user_index{lc $p_user});
  }

  # For users updating their own accounts, the old password must match the current password.
  if ($is_own_acc && $query->param('pass') && ($user_data[$user_num]{'pass'} ne
   crypt($query->param('oldpass'), $user_data[$user_num]{'pass'})))
  {
   $acc_error .= "<div>Your old password is incorrect.</div>\n";
  }

  # For all form submittals, the two passwords must be either both nothing, or both identical.
  if (($is_new_acc && !$p_pass) || ($p_pass ne $p_pass2))
  {
   $acc_error .= "<div>Please ensure your new passwords are identical!</div>\n";
  }

  # Check the location/bio/sig lines against maximum permissable length.
  if (length $query->param('location') > 30)
  {
   $acc_error .= "<div>Your 'Location' is too long. The maximum length is 30 characters, yours is ".
    "currently ". (length $query->param('location')) ." characters.</div>\n";
  }
  if (length $query->param('bio') > 1000)
  {
   $acc_error .= "<div>Your 'Bio' is too long. The maximum length is 1000 characters, yours is ".
    "currently ". (length $query->param('bio')) ." characters.</div>\n";
  }
  if (length $query->param('sig') > 200)
  {
   $acc_error .= "<div>Your 'Sig' is too long. The maximum length is 200 characters, yours is ".
    "currently ". (length $query->param('sig')) ." characters.</div>\n";
  }

  # Spambot checks
  if ($query->param('email'))
  {
   $spam_error = 1;
  }

  # Check the submitted timestamp isn't within the last few seconds (non-admins)
  if (!$query->param('ts'))
  {
   $spam_error = 1;
  }
  elsif (!$user_priv{'anyedit'})
  {
   for (my $i = 0; $i < 6; $i++)
   {
    my $timestamp = crypt(reverse(time() - $i), "TS");
    if ($query->param('ts') eq $timestamp)
    {
     $acc_error .= "<div>Slow down - you are posting too fast!</div>\n";
     last;
    }
   }
  }

  # Otherwise jump past this straight to displaying the form so it can be amended.
  if (!$acc_error)
  {
   # First, come up with the account name & number we're going to modify.
   if ($is_new_acc)
   {
    # Decide upon a new user number if it's a new account.
    $acc_num = 1;
    foreach my $old_num (values %user_index)
    {
     $acc_num = $old_num + 1 if $old_num >= $acc_num;
    }
    $acc_name = $p_user;
   }
   # Either create or update the user entry in @user_data.
   if (!$user_data[$acc_num])
   {
    # Set permissions... the first user is an administrator, the rest are normal members.
    $user_data[$acc_num] = {
	 'joined' => time,
	 'posts' => 0,
	 'name' => $acc_name,
	 'group' => ($acc_num == 1) ? 'admins' : 'members'
    };
   }
   # Admins are allowed to edit others' membership in usergroups.
   if ($user_priv{'accedit'})
   {
    my $gs = $query->param('groupselect');
    $user_data[$acc_num]{'group'} = $gs if ($gs);
   }
   if ($p_pass)
   {
    # Update the user's stored & encrypted password.
    $user_data[$acc_num]{'pass'} = crypt($p_pass, join('',('A'..'Z','a'..'z',0..9)[rand 62,rand 62])),
   }
   # And store the rest of the data.
   $user_data[$acc_num]{'location'} = $location;
   $user_data[$acc_num]{'bio'} = $bio;
   $user_data[$acc_num]{'sig'} = $sig;

   # Check that the user data folder exists.
   if (!(-e $forum_prefs{'user_dir'}))
   {
    mkdir("$forum_prefs{'user_dir'}", 0755) or die "Couldn't create user data folder, $!";
   }
   # Update our user index to reflect this new user.
   $user_index{lc $acc_name} = $acc_num;
   if (!$spam_error)
   {
    writeFileData("$forum_prefs{'user_dir'}/index.txt", [\%user_index]);
   }
   # If it's a new account or password change, userLogin() will set cookies and write @user_data.
   # Otherwise, if we're just updating an existing account, we shouldn't record this as a login.
   if ($is_new_acc || ($p_pass && $is_own_acc))
   {
    userLogin();
   }
   elsif (!$spam_error)
   {
    writeFileData("$forum_prefs{'user_dir'}/$acc_num.txt", [$user_data[$acc_num]]);
   }

   # For new accounts, send them back to the frontpage to get them started ;).
   if ($is_new_acc)
   {
    return listForums();
   }

   # Otherwise, tell the user how industrious we've been.
   $acc_error = "<div>Account changes saved!</div>\n";
  }
 }


 # Finally, whatever happens, write the account info form! And any corrected values we have.
 # For logged-in users who haven't posted form data, retrieve their current values.
 if ($acc_num)
 {
  $location |= $user_data[$acc_num]{'location'};
  $bio |= $user_data[$acc_num]{'bio'};
  $sig |= $user_data[$acc_num]{'sig'};
 }
 # Compile the account type select box.
 my $groupselect = '';
 foreach my $g (keys %user_groups)
 {
  $groupselect .= " <option value=\"$g\"".
   ($g eq $user_data[$acc_num]{'group'} ? ' selected="selected"' : '').
   ">$g ($user_groups{$g}{'name'})</option>\n";
 }

 # And display whatever we've got.
 printHeader('Account Setup');
 getTemplate('userAccountEdit');
 s/\[\?ACCEDIT(.)(.*?)\1(.*?)\?\]/$user_priv{'accedit'} ? $2 : $3/seg;
 doSub('GROUPSELECT', $groupselect);
 doSub('ACCOUNTPOSTURL', $forum_prefs{'path_name'});
 doSub('ACCOUNTERROR', $p_user ? $acc_error : "");
 doSub('ACCOUNTUSER', url_dec($p_user||$user_name));
 doSub('HIDDENFIELDS', $hiddenfields);
 $bio =~ s/<br \/>/\n/g;
 $sig =~ s/<br \/>/\n/g;
 doSub('LOCATION', $location);
 doSub('BIO', $bio);
 doSub('SIG', $sig);
 printTemplate();
 printFooter();
}


sub userAccountDelete
{
 # Not used anywhere else, can call manually. Pass a user name.
 my $acc_name = lc $_[0];
 my $acc_num = $user_index{$acc_name};
 $acc_num or return;

 if (!$user_data[$acc_num])
 {
  my @tmp = readFileData("$forum_prefs{'user_dir'}/$acc_num.txt");
  $user_data[$acc_num] = \%{$tmp[0]};
  $user_data[$acc_num] or return;
 }
 # TODO: Can't delete users that have posts still in the forum.
 $user_data[$acc_num]{'posts'} and return;

 delete $user_index{$acc_name};
 writeFileData("$forum_prefs{'user_dir'}/index.txt", [\%user_index]);
 splice(@user_data, $acc_num, 1);
 unlink "$forum_prefs{'base_dir'}/_users/$acc_num.txt";
}









##### MAIN SCRIPT ENTRY POINT #####


# Get global script parameters from the path info or URL query.

$forum_prefs{'path_name'} |= '';
my $path_str = $ENV{'PATH_INFO'} || '';
$path_str =~ s/^\Q$forum_prefs{'path_name'}\E//i;
@path_info = $path_str ? split(/\//, $path_str) : ();

$forum = ($query->param('f') || $query->param('forum') || $path_info[1] || '');
$forum = ($forum =~ m/^([A-Za-z0-9]{1,50})/ ? $1 : '');

$topic = ($query->param('t') || $query->param('topic') || $path_info[2] || '');
$topic = ($topic =~ m/^([0-9]{1,6})/ ? $1 : 0);

$act = ($query->param('a') || $query->param('act') || '');

$page = ($query->param('p') || ($path_str =~ m/p\=([0-9]+)/ ? $1 : 0)) || 1;

# Limit keywords to 100 chars, trim trailing/leading spaces.
$keywords = ($query->param('keywords') =~ m/^\s*(.+?)\s*$/) ? trimHTML($1) : "";
$keywords =~ s/^(.{1,100}).*/$1/g;




# Now check for the existence of the FORUMS.TXT and TEMPLATE.PL files.
# Generate an error if they can't be found.
if (!(-e "$forum_prefs{'base_dir'}/forums.txt") ||
 !(-e "$forum_prefs{'base_dir'}/template.pl"))
{
 print qq|Content-type: text/html

<html><body>
<h2>TwinHelix Forums: Configuration</h2>
<div>Congratulations, your Perl interpreter is correctly running this script!
Next, you need to configure your data files directory.</div>
<div>Please make sure, in this main Perl script, that the \$forum_prefs{'base_dir'} variable
points to a folder containing the files "forum.txt" and "template.pl".</div>
<div>These cannot be located at present, the script is looking for them in:</div>
<div>$forum_prefs{'base_dir'}</div>
</body></html>
|;
}
elsif (!(do "$forum_prefs{'base_dir'}/template.pl"))
{
 # Display an error if the template file can't be evaluated successfully.
 print qq|Content-type: text/html

<html><body>
<h2>TwinHelix Forums: Error</h2>
<div>There's an error in your TEMPLATE file:</div>
<div>$@</div>
</body></html>
|;
}
else
{
 # Work out what's banned from this forum.
 my $host_banned = 0;
 foreach my $ban_env (keys %forum_banned)
 {
  foreach my $ban_entry (@{$forum_banned{$ban_env}})
  {
   if (!$host_banned && (($ENV{$ban_env}||'') =~ m/(\Q$ban_entry\E)/i))
   {
    printHeader('Forum Access Disallowed');
    forum_out("<h1>Forum Access Disallowed</h1>\n");
    forum_out("<div>Access is not allowed from $ban_env: $1.</div>\n");
    printFooter();
    $host_banned = 1;
   }
  }
 }

 if (!$host_banned)
 {
  # We must be good to go from here on.

  # Read an array @forum_data that contains hashes of forum names/descriptions.
  @forum_data = readFileData("$forum_prefs{'base_dir'}/forums.txt");

  # Check the query's $forum against the list from disk...
  # Find the correct reference in the array of forums read earlier.
  for (my $i = 0; $i <= $#forum_data; $i++)
  {
   if (($forum_data[$i]{'id'}||'') eq $forum) { $forum_ref = \%{$forum_data[$i]} }
  }
  $forum = '' if (!$forum_ref);


  # Now pull our user data into our hash...
  my @tmp = readFileData("$forum_prefs{'user_dir'}/index.txt");
  %user_index = %{$tmp[0]} if ($#tmp > -1);
  # Eat cookie! :). Find & decode the provided username and corresponding number.
  my $cookie_str = $ENV{'HTTP_COOKIE'} || $ENV{'COOKIE'} || '';
  my $uname = '';
  my $sessid = '';
  if ($cookie_str =~ m/user=([^\.]+).([^;]+)/)
  {
   $uname = url_dec($1);
   $sessid = $2;
  }
  my $unum = $user_index{lc $uname};

  if ($unum)
  {
   # Retrieve the info for this username from the appropriate numbered filename.
   @tmp = readFileData("$forum_prefs{'user_dir'}/$unum.txt");
   $user_data[$unum] = \%{$tmp[0]};
   # Pull out the /16 subnets of the stored and current IP addresses.
   # For IPV6 just hack this **TODO**
   my $last_ip = $1 if (($user_data[$unum]{'ip'}||'') =~ m/^\W*(\w+[:|\.]\w+)/);
   my $curr_ip = $1 if (($ENV{'REMOTE_ADDR'}||'') =~ m/^\W*(\w+[:|\.]\w+)/);
   # Finally, check the remote session ID/IP address/timeout against the stored values.
   # If they're all fine, set our global user_name and user_num variables, we're logged in.
   if (($sessid && $sessid eq $user_data[$unum]{'sessionid'}) &&
	($last_ip && ($last_ip eq $curr_ip)) &&
	(time < $user_data[$unum]{'lastlogin'} + $forum_prefs{'session_length'}*60*60))
   {
    $user_name = $user_data[$unum]{'name'};
    $user_num = $unum;
   }
  }
  # And record the privilege group of this current user.
  %user_priv = %{$user_groups{$user_num?$user_data[$user_num]{'group'}:'guests'}};


  # Depending on CGI parameters, display the content.
  if ($act eq 'login') { userLogin() }
  elsif ($act eq 'logout' || $path_info[1] eq '_logout') { userLogout() }
  elsif ($act eq 'users' || $path_info[1] eq '_user') { userAccountView() }
  elsif ($act eq 'account' || $path_info[1] eq '_account') { userAccountEdit() }
  elsif (!$forum) { listForums() }
  elsif ($act eq 'newtopic' || $path_info[2] eq '_newtopic') { newTopic() }
  elsif ($act eq 'postmessage') { postMessage() }
  elsif (!$topic) { listTopics() }
  elsif ($act eq 'edit' || $path_info[3] =~ /^edit=/) { editPost() }
  elsif ($act eq 'topicsettings') { processTopicSettings() }
  else { displayTopic() }
 }
}


# BONUS: Delete spammer accounts, uncomment to use. Will thrash the disk.
#my $name, $num, %delusers;
#while (($name, $num) = each %user_index)
#{
# my @tmp = readFileData("$forum_prefs{'user_dir'}/$num.txt");
# $user_data[$num] = \%{$tmp[0]};
# $user_data[$num] or next;
# if ($num gt 1000 && !$user_data[$num]{'posts'} && (
#   ($user_data[$num]{'lastlogin'} - $user_data[$num]{'joined'} lt 60) ||
#   ($user_data[$num]{'group'} eq 'banned')
#  ))
# {
#  $delusers{$name} = $num;
# }
#}
#while (($name, $num) = each %delusers)
#{
# userAccountDelete($name);
#}
